package com.jg.common;

import io.swagger.annotations.ApiModel;
import io.swagger.annotations.ApiModelProperty;
import lombok.Data;

import java.io.Serializable;
import java.util.Optional;

/**
 * <p>
 *     全局统一返回结果
 * </p>
 *
 * @author lxp
 * @version V1.0
 * @since 2022/7/19  23:43
 */
@Data
@ApiModel(value = "全局统一返回结果")
public class Result<T> implements Serializable {

    private static final long serialVersionUID = 1L;

    @ApiModelProperty(value = "业务错误码")
    private long code;

    @ApiModelProperty(value = "返回数据")
    private T data;

    @ApiModelProperty(value = "返回消息")
    private String message;

    private Result(){
        // to do nothing
    }

    public Result(ErrorCode errorCode) {
        errorCode = Optional.ofNullable(errorCode).orElse(ApiErrorCode.FAILED);
        this.code = errorCode.getCode();
        this.message = errorCode.getMessage();
    }

    public static <T> Result<T> ok() {
        return restResult(null, ApiErrorCode.SUCCESS);
    }

    public static <T> Result<T> ok(T data) {
        ApiErrorCode aec = ApiErrorCode.SUCCESS;
        if (data instanceof Boolean && Boolean.FALSE.equals(data)) {
            aec = ApiErrorCode.FAILED;
        }
        return restResult(data, aec);
    }

    public static <T> Result<T> failed(String message) {
        return restResult(null, ApiErrorCode.FAILED.getCode(), message);
    }

    public static <T> Result<T> failed(ErrorCode errorCode) {
        return restResult(null, errorCode);
    }

    public static <T> Result<T> failed(long errorCode, String message) {
        return restResult(null, errorCode, message);
    }

    public static <T> Result<T> failed(ErrorCode errorCode, String message) {
        return restResult(null, errorCode.getCode(), message);
    }

    public static <T> Result<T> restResult(T data, ErrorCode errorCode) {
        return restResult(data, errorCode.getCode(), errorCode.getMessage());
    }

    private static <T> Result<T> restResult(T data, long code, String message) {
        Result<T> apiResult = new Result<>();
        apiResult.setCode(code);
        apiResult.setData(data);
        apiResult.setMessage(message);
        return apiResult;
    }
}
