package com.jg.service;

import com.intelligt.modbus.jlibmodbus.data.DataHolder;
import com.intelligt.modbus.jlibmodbus.exception.IllegalDataAddressException;
import com.intelligt.modbus.jlibmodbus.exception.IllegalDataValueException;

import java.util.ArrayList;
import java.util.List;

/**
 * <p>
 *     功能码
 *          0x01: 读线圈寄存器
 *          0x02: 读离散输入寄存器
 *          0x03: 读保持寄存器
 *          0x04: 读输入寄存器
 *          0x05: 写单个线圈寄存器
 *          0x06: 写单个保持寄存器
 *          0x0f: 写多个线圈寄存器
 *          0x10: 写多个保持寄存器
 * </p>
 *
 * @author lxp
 * @version V1.0
 * @since 2023/6/2  15:31
 */
public class MyOwnDataHolder extends DataHolder {

    final List<ModBusEventListener> modBusEventListenerList = new ArrayList<>();

    public MyOwnDataHolder() {
        // you can place the initialization code here
        /*
         * something like that: setHoldingRegisters(new
         * SimpleHoldingRegisters(10)); setCoils(new Coils(128)); ... etc.
         */
    }

    public void addEventListener(ModBusEventListener listener) {
        modBusEventListenerList.add(listener);
    }

    public boolean removeEventListener(ModBusEventListener listener) {
        return modBusEventListenerList.remove(listener);
    }

    @Override
    public int[] readHoldingRegisterRange(int offset, int quantity) throws IllegalDataAddressException {
        for (ModBusEventListener l : modBusEventListenerList) {
            l.readHoldingRegisterRange(offset, quantity);
        }
        return super.readHoldingRegisterRange(offset, quantity);
    }

    @Override
    public int[] readInputRegisterRange(int offset, int quantity) throws IllegalDataAddressException{
        for (ModBusEventListener l : modBusEventListenerList) {
            l.readInputRegisterRange(offset, quantity);
        }
        return super.readInputRegisterRange(offset, quantity);
    }

    @Override
    public void writeHoldingRegister(int offset, int value) throws IllegalDataAddressException,
            IllegalDataValueException {
        for (ModBusEventListener l : modBusEventListenerList) {
            l.onWriteToSingleHoldingRegister(offset, value);
        }
        super.writeHoldingRegister(offset, value);
    }

    @Override
    public void writeHoldingRegisterRange(int offset, int[] range)
            throws IllegalDataAddressException, IllegalDataValueException {
        for (ModBusEventListener l : modBusEventListenerList) {
            l.onWriteToMultipleHoldingRegisters(offset, range.length, range);
        }
        super.writeHoldingRegisterRange(offset, range);
    }

    @Override
    public void writeCoil(int offset, boolean value) throws IllegalDataAddressException,
            IllegalDataValueException {
        for (ModBusEventListener l : modBusEventListenerList) {
            l.onWriteToSingleCoil(offset, value);
        }
        super.writeCoil(offset, value);
    }

    @Override
    public void writeCoilRange(int offset, boolean[] range) throws IllegalDataAddressException,
            IllegalDataValueException {
        for (ModBusEventListener l : modBusEventListenerList) {
            l.onWriteToMultipleCoils(offset, range.length, range);
        }
        super.writeCoilRange(offset, range);
    }

}
